#include "sh.local.h"
/*
 * C shell
 */

extern  int await();
extern  int chngd();
extern  int doalias();
extern  int dobreak();
extern  int docontin();
extern  int doecho();
extern  int doelse();
extern  int doend();
extern  int doendif();
extern  int doendsw();
extern  int doexit();
extern  int doforeach();
extern  int doglob();
#ifdef  GOGO
extern  int dogogo();
#endif  GOGO
extern  int dogoto();
extern  int dohash();
extern  int dohist();
extern  int doif();
extern  int dolet();
extern  int dologin();
extern  int dologout();
extern  int donewgrp();
extern  int donice();
extern  int donohup();
extern  int doonintr();
extern  int dorepeat();
extern  int doset();
extern  int dosetenv();
extern  int dosource();
extern  int doswbrk();
extern  int doswitch();
extern  int dotime();
#ifndef V6
extern  int doumask();
#endif
extern  int dowhile();
extern  int dozip();
extern  int execash();
#ifdef VFORK
extern  int hashstat();
#endif
extern  int goodbye();
extern  int shift();
extern  int showall();
extern  int unalias();
extern  int dounhash();
extern  int unset();
extern  int doeval();

#define INF     1000

struct  biltins {
	char    *bname;
	int     (*bfunct)();
	short   minargs, maxargs;
} bfunc[] = {
	"@",            dolet,          0,      INF,
	"alias",        doalias,        0,      INF,
#ifdef debug
	"alloc",        showall,        0,      1,
#endif
	"break",        dobreak,        0,      0,
	"breaksw",      doswbrk,        0,      0,
	"case",         dozip,          0,      1,
	"cd",           chngd,          0,      1,
	"chdir",        chngd,          0,      1,
	"continue",     docontin,       0,      0,
	"default",      dozip,          0,      0,
	"echo",         doecho,         0,      INF,
	"else",         doelse,         0,      INF,
	"end",          doend,          0,      0,
	"endif",        dozip,          0,      0,
	"endsw",        dozip,          0,      0,
	"exec",         execash,        1,      INF,
	"exit",         doexit,         0,      INF,
	"eval",         doeval,         0,      INF,
	"foreach",      doforeach,      3,      INF,
	"glob",         doglob,         0,      INF,
#ifdef  GOGO
	"gogo",         dogogo,         1,      1,
#endif  GOGO
	"goto",         dogoto,         1,      1,
#ifdef VFORK
	"hashstat",     hashstat,       0,      0,
#endif
	"history",      dohist,         0,      0,
	"if",           doif,           1,      INF,
	"login",        dologin,        0,      1,
	"logout",       dologout,       0,      0,
	"newgrp",       donewgrp,       1,      1,
	"nice",         donice,         0,      INF,
	"nohup",        donohup,        0,      INF,
	"onintr",       doonintr,       0,      2,
	"rehash",       dohash,         0,      0,
	"repeat",       dorepeat,       2,      INF,
	"set",          doset,          0,      INF,
#ifndef V6
	"setenv",       dosetenv,       0,      2,
#endif
	"shift",        shift,          0,      1,
	"source",       dosource,       1,      1,
	"switch",       doswitch,       1,      INF,
	"time",         dotime,         0,      INF,
#ifndef V6
	"umask",        doumask,        0,      1,
#endif
	"unalias",      unalias,        1,      INF,
	"unhash",       dounhash,               0,      0,
	"unset",        unset,          1,      INF,
	"wait",         await,          0,      0,
	"while",        dowhile,        1,      INF,
	0,              0,              0,      0,
};

#define ZBREAK          0
#define ZBRKSW          1
#define ZCASE           2
#define ZDEFAULT        3
#define ZELSE           4
#define ZEND            5
#define ZENDIF          6
#define ZENDSW          7
#define ZEXIT           8
#define ZFOREACH        9
#define ZGOTO           10
#define ZIF             11
#define ZLABEL          12
#define ZLET            13
#define ZSET            14
#define ZSWITCH         15
#define ZTEST           16
#define ZTHEN           17
#define ZWHILE          18

struct srch {
	char    *s_name;
	short   s_value;
} srchn[] = {
	"@",            ZLET,
	"break",        ZBREAK,
	"breaksw",      ZBRKSW,
	"case",         ZCASE,
	"default",      ZDEFAULT,
	"else",         ZELSE,
	"end",          ZEND,
	"endif",        ZENDIF,
	"endsw",        ZENDSW,
	"exit",         ZEXIT,
	"foreach",      ZFOREACH,
	"goto",         ZGOTO,
	"if",           ZIF,
	"label",        ZLABEL,
	"set",          ZSET,
	"switch",       ZSWITCH,
	"while",        ZWHILE,
	0,              0,
};

char    *mesg[] = {
	0,
	"Hangup",
	0,
	"Quit",
	"Illegal instruction",
	"Trace/BPT trap",
	"IOT trap",
	"EMT trap",
	"Floating exception",
	"Killed",
	"Bus error",
	"Segmentation violation",
	"Bad system call",
	0,
	"Alarm clock",
	"Terminated"
};

#ifdef EDIAG

char    *Rmesg[] = {
	0,
	"",
	0,
	"",
	" ",
	"  trace/BPT",
	"  IOT",
	"  EMT",
	" . ",
	"",
	" ",
	"  ",
	"  ",
	0,
	" alarm",
	""
};

#endif EDIAG

/*
 *   / 
 */

char to8quot[256]= {
/* 0   1   2    3    4    5    6    7  */
 0200,0201,0202,0203,0204,0205,0206,0207,
/*  10   11   12   13   14   15   16   17 */
  0210,0211,0212,0213,0214,0215,0216,0217,
/*  20   21   22   23   24   25   26   27 */
  0220,0221,0222,0223,0224,0225,0226,0227,
/*  30   31   32   33   34   35   36   37 */
  0230,0231,0232,0233,0234,0235,0236,0237,
/*
   :40 !:41 ":42 #:43 $:44 %:45 &:46 ':47 */
  0240,0241,0242,0243,0244,0245,0246,0247,
/*(:50 ):51 *:52 +:53 ,:54 -:55 .:56 /:57 */
  0250,0251,0252,0253,0254,0255,0256,0257,
/*0:60 1:61 2:62 3:63 4:64 5:65 6:66 7:67 -    */
  0060,0061,0062,0063,0064,0065,0066,0067,
/*8:70 9:71 ::72 ;:73 <:74 =:75 >:76 ?:77 */
  0070,0071,0272,0273,0274,0275,0276,0277,
/*
  @100 A101 B102 C103 D104 E105 F106 G107 */
  0260,0101,0102,0103,0104,0105,0106,0107,
/*H110 I111 J112 K113 L114 M115 N116 O117 */
  0110,0111,0112,0113,0114,0115,0116,0117,
/*P120 Q121 R122 S123 T124 U125 V126 W127 */
  0120,0121,0122,0123,0124,0125,0126,0127,
/*X130 Y131 Z132 [133 \134 ]135 ^136 _137 */
  0130,0131,0132,0261,0262,0263,0264,0137,
/*
  `140 a141 b142 c143 d144 e145 f146 g147 */
  0265,0141,0142,0143,0144,0145,0146,0147,
/*h150 i151 j152 k153 l154 m155 n156 o157 */
  0150,0151,0152,0153,0154,0155,0156,0157,
/*p160 q161 r162 s163 t164 u165 v166 w167 */
  0160,0161,0162,0163,0164,0165,0166,0167,
/*x170 y171 z172 {173 |174 }175 ~176  177 */
#ifndef MHOC
  0170,0171,0172,0266,0267,0270,0271,0377,
#else
  0170,0171,0172,0266,0267,0270,0271,0177,
#endif
/*      */
  0200,0201,0202,0203,0204,0205,0206,0207,
  0210,0211,0212,0213,0214,0215,0216,0217,
  0220,0221,0222,0223,0224,0225,0226,0227,
  0230,0231,0232,0233,0234,0235,0236,0237,
  0240,0241,0242,0243,0244,0245,0246,0247,
  0250,0251,0252,0253,0254,0255,0256,0257,
  0260,0261,0262,0263,0264,0265,0266,0267,
  0270,0271,0272,0273,0274,0275,0276,0277,
  0300,0301,0302,0303,0304,0305,0306,0307,
  0310,0311,0312,0313,0314,0315,0316,0317,
  0320,0321,0322,0323,0324,0325,0326,0327,
  0330,0331,0332,0333,0334,0335,0336,0337,
  0340,0341,0342,0343,0344,0345,0346,0347,
  0350,0351,0352,0353,0354,0355,0356,0357,
  0360,0361,0362,0363,0364,0365,0366,0367,
  0370,0371,0372,0373,0374,0375,0376,0377 };

char to8trim[256]= {
/*      */
  0000,0001,0002,0003,0004,0005,0006,0007,
  0010,0011,0012,0013,0014,0015,0016,0017,
  0020,0021,0022,0023,0024,0025,0026,0027,
  0030,0031,0032,0033,0034,0035,0036,0037,
  0040,0041,0042,0043,0044,0045,0046,0047,
  0050,0051,0052,0053,0054,0055,0056,0057,
  0060,0061,0062,0063,0064,0065,0066,0067,
  0070,0071,0072,0073,0074,0075,0076,0077,
  0100,0101,0102,0103,0104,0105,0106,0107,
  0110,0111,0112,0113,0114,0115,0116,0117,
  0120,0121,0122,0123,0124,0125,0126,0127,
  0130,0131,0132,0133,0134,0135,0136,0137,
  0140,0141,0142,0143,0144,0145,0146,0147,
  0150,0151,0152,0153,0154,0155,0156,0157,
  0160,0161,0162,0163,0164,0165,0166,0167,
  0170,0171,0172,0173,0174,0175,0176,0177,
/* .   */
  0000,0001,0002,0003,0004,0005,0006,0007,
  0010,0011,0012,0013,0014,0015,0016,0017,
  0020,0021,0022,0023,0024,0025,0026,0027,
  0030,0031,0032,0033,0034,0035,0036,0037,
/*        */
  0040,0041,0042,0043,0044,0045,0046,0047,
  0050,0051,0052,0053,0054,0055,0056,0057,
  0100,0133,0134,0135,0136,0140,0173,0174,  /*   -   */
  0175,0176,0072,0073,0074,0075,0076,0077,
/*     */
  0300,0301,0302,0303,0304,0305,0306,0307,
  0310,0311,0312,0313,0314,0315,0316,0317,
  0320,0321,0322,0323,0324,0325,0326,0327,
  0330,0331,0332,0333,0334,0335,0336,0337,
  0340,0341,0342,0343,0344,0345,0346,0347,
/*     */
  0350,0351,0352,0353,0354,0355,0356,0357,
  0360,0361,0362,0363,0364,0365,0366,0367,
#ifndef MHOC
  0370,0371,0372,0373,0374,0375,0376,0177 };
#else
  0370,0371,0372,0373,0374,0375,0376,0377 };
#endif

char *toquot = to8quot, *totrim = to8trim; /*    2-  */
